/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*!
 * \file
 * \brief RADIUS CDR Support
 *
 * \author Philippe Sultan
 * The Radius Client Library
 * 	http://developer.berlios.de/projects/radiusclient-ng/
 *
 * \arg See also \ref AstCDR
 * \ingroup cdr_drivers
 */

/*! \li \ref cdr_radius.c uses the configuration file \ref cdr.conf
 * \addtogroup configuration_file Configuration Files
 */

/*** MODULEINFO
	<depend>radius</depend>
	<support_level>extended</support_level>
 ***/

#include "asterisk.h"

#include RADIUS_HEADER_STR

#include "asterisk/channel.h"
#include "asterisk/cdr.h"
#include "asterisk/module.h"
#include "asterisk/utils.h"

/*! ISO 8601 standard format */
#define DATE_FORMAT "%Y-%m-%d %T %z"

#define VENDOR_CODE           22736

enum {
	PW_AST_ACCT_CODE =    101,
	PW_AST_SRC =          102,
	PW_AST_DST =          103,
	PW_AST_DST_CTX =      104,
	PW_AST_CLID =         105,
	PW_AST_CHAN =         106,
	PW_AST_DST_CHAN =     107,
	PW_AST_LAST_APP =     108,
	PW_AST_LAST_DATA =    109,
	PW_AST_START_TIME =   110,
	PW_AST_ANSWER_TIME =  111,
	PW_AST_END_TIME =     112,
	PW_AST_DURATION =     113,
	PW_AST_BILL_SEC =     114,
	PW_AST_DISPOSITION =  115,
	PW_AST_AMA_FLAGS =    116,
	PW_AST_UNIQUE_ID =    117,
	PW_AST_USER_FIELD =   118,
	PW_AST_USER_SESSION_ID =   119,
	PW_AST_USER_SESSION_ID2 =   120,
	PW_AST_SIP_IP_ADDR =   121,
	PW_AST_FAR_END_IP_ADDR =   122,
	PW_AST_SIP_RESPONSE_CODE =   123,
	PW_AST_CODEC =   124,
	PW_AST_RTP_LOCAL_BURST_DENSITY =   125,
	PW_AST_RTP_REMOTE_BURST_DENSITY =   126,
	PW_AST_RTP_LOCAL_BURST_DURATION =   127,
	PW_AST_RTP_REMOTE_BURST_DURATION =   128,
	PW_AST_RTP_LOCAL_GAP_DENSITY =   129,
	PW_AST_RTP_REMOTE_GAP_DENSITY =   130,
	PW_AST_RTP_LOCAL_GAP_DURATION =   131,
	PW_AST_RTP_REMOTE_GAP_DURATION =   132,
	PW_AST_RTP_LOCAL_JB_RATE =   133,
	PW_AST_RTP_REMOTE_JB_RATE =   134,
	PW_AST_RTP_LOCAL_JB_MAX =   135,
	PW_AST_RTP_REMOTE_JB_MAX =   136,
	PW_AST_RTP_LOCAL_JB_NOMINAL =   137,
	PW_AST_RTP_REMOTE_JB_NOMINAL =   138,
	PW_AST_RTP_LOCAL_JB_ABS_MAX =   139,
	PW_AST_RTP_REMOTE_JB_ABS_MAX =   140,
	PW_AST_RTP_JB_AVG =   141,
	PW_AST_RTP_LOSS_RATE =   142,
	PW_AST_RTP_DISCARDED =   143,
	PW_AST_RTP_LOST =   144,
	PW_AST_RTP_RX_PKTS =   145,
	PW_AST_RTP_TX_PKTS =   146,
	PW_AST_RTP_JITTER =   147,
	PW_AST_RTP_MAX_JITTER =   148,
	PW_AST_RTP_AVERAGE_ROUND_TRIP_DELAY =   149,
	PW_AST_RTP_AVERAGE_FAR_END_INTERARRIVAL_JITTER =   150
};

enum {
	/*! Log dates and times in UTC */
	RADIUS_FLAG_USEGMTIME = (1 << 0),
	/*! Log Unique ID */
	RADIUS_FLAG_LOGUNIQUEID = (1 << 1),
	/*! Log User Field */
	RADIUS_FLAG_LOGUSERFIELD = (1 << 2)
};

static const char desc[] = "RADIUS CDR Backend";
static const char name[] = "radius";
static const char cdr_config[] = "cdr.conf";

#ifdef FREERADIUS_CLIENT
static char radiuscfg[PATH_MAX] = "/etc/radiusclient/radiusclient.conf";
#else
static char radiuscfg[PATH_MAX] = "/etc/radiusclient-ng/radiusclient.conf";
#endif

static struct ast_flags global_flags = { RADIUS_FLAG_USEGMTIME | RADIUS_FLAG_LOGUNIQUEID | RADIUS_FLAG_LOGUSERFIELD };

static rc_handle *rh = NULL;

static int build_radius_record(VALUE_PAIR **tosend, struct ast_cdr *cdr)
{
	int recordtype = PW_STATUS_STOP;
	struct ast_tm tm;
	char timestr[128];
	char *tmp;

	if (!rc_avpair_add(rh, tosend, PW_ACCT_STATUS_TYPE, &recordtype, 0, 0))
		return -1;

	/* Account code */
	if (!rc_avpair_add(rh, tosend, PW_AST_ACCT_CODE, &cdr->accountcode, strlen(cdr->accountcode), VENDOR_CODE))
		return -1;

 	/* Source */
	if (!rc_avpair_add(rh, tosend, PW_AST_SRC, &cdr->src, strlen(cdr->src), VENDOR_CODE))
		return -1;

 	/* Destination */
	if (!rc_avpair_add(rh, tosend, PW_AST_DST, &cdr->dst, strlen(cdr->dst), VENDOR_CODE))
		return -1;

 	/* Destination context */
	if (!rc_avpair_add(rh, tosend, PW_AST_DST_CTX, &cdr->dcontext, strlen(cdr->dcontext), VENDOR_CODE))
		return -1;

	/* Caller ID */
	if (!rc_avpair_add(rh, tosend, PW_AST_CLID, &cdr->clid, strlen(cdr->clid), VENDOR_CODE))
		return -1;

	/* Channel */
	if (!rc_avpair_add(rh, tosend, PW_AST_CHAN, &cdr->channel, strlen(cdr->channel), VENDOR_CODE))
		return -1;

	/* Destination Channel */
	if (!rc_avpair_add(rh, tosend, PW_AST_DST_CHAN, &cdr->dstchannel, strlen(cdr->dstchannel), VENDOR_CODE))
		return -1;

	/* Last Application */
	if (!rc_avpair_add(rh, tosend, PW_AST_LAST_APP, &cdr->lastapp, strlen(cdr->lastapp), VENDOR_CODE))
		return -1;

	/* Last Data */
	if (!rc_avpair_add(rh, tosend, PW_AST_LAST_DATA, &cdr->lastdata, strlen(cdr->lastdata), VENDOR_CODE))
		return -1;


	/* Start Time */
	ast_strftime(timestr, sizeof(timestr), DATE_FORMAT,
		ast_localtime(&cdr->start, &tm,
			ast_test_flag(&global_flags, RADIUS_FLAG_USEGMTIME) ? "GMT" : NULL));
	if (!rc_avpair_add(rh, tosend, PW_AST_START_TIME, timestr, strlen(timestr), VENDOR_CODE))
		return -1;

	/* Answer Time */
	ast_strftime(timestr, sizeof(timestr), DATE_FORMAT,
		ast_localtime(&cdr->answer, &tm,
			ast_test_flag(&global_flags, RADIUS_FLAG_USEGMTIME) ? "GMT" : NULL));
	if (!rc_avpair_add(rh, tosend, PW_AST_ANSWER_TIME, timestr, strlen(timestr), VENDOR_CODE))
		return -1;

	/* End Time */
	ast_strftime(timestr, sizeof(timestr), DATE_FORMAT,
		ast_localtime(&cdr->end, &tm,
			ast_test_flag(&global_flags, RADIUS_FLAG_USEGMTIME) ? "GMT" : NULL));
	if (!rc_avpair_add(rh, tosend, PW_AST_END_TIME, timestr, strlen(timestr), VENDOR_CODE))
		return -1;

 	/* Duration */
	if (!rc_avpair_add(rh, tosend, PW_AST_DURATION, &cdr->duration, 0, VENDOR_CODE))
		return -1;

	/* Billable seconds */
	if (!rc_avpair_add(rh, tosend, PW_AST_BILL_SEC, &cdr->billsec, 0, VENDOR_CODE))
		return -1;

	/* Disposition */
	tmp = ast_strdupa(ast_cdr_disp2str(cdr->disposition));
	if (!rc_avpair_add(rh, tosend, PW_AST_DISPOSITION, tmp, strlen(tmp), VENDOR_CODE))
		return -1;

	/* AMA Flags */
	tmp = ast_strdupa(ast_channel_amaflags2string(cdr->amaflags));
	if (!rc_avpair_add(rh, tosend, PW_AST_AMA_FLAGS, tmp, strlen(tmp), VENDOR_CODE))
		return -1;

	if (ast_test_flag(&global_flags, RADIUS_FLAG_LOGUNIQUEID)) {
		/* Unique ID */
		if (!rc_avpair_add(rh, tosend, PW_AST_UNIQUE_ID, &cdr->uniqueid, strlen(cdr->uniqueid), VENDOR_CODE))
			return -1;
	}

	if (ast_test_flag(&global_flags, RADIUS_FLAG_LOGUSERFIELD)) {
		/* append the user field */
		if (!rc_avpair_add(rh, tosend, PW_AST_USER_FIELD, &cdr->userfield, strlen(cdr->userfield), VENDOR_CODE))
			return -1;
	}

	/* SessionId */
	if (!rc_avpair_add(rh, tosend, PW_AST_USER_SESSION_ID, &cdr->sessionId, 0, VENDOR_CODE))
		return -1;

	/* SIPSessionID */
	if (!rc_avpair_add(rh, tosend, PW_AST_USER_SESSION_ID2, &cdr->SIPSessionID, strlen(cdr->SIPSessionID), VENDOR_CODE))
		return -1;

	/* SIP IP Address */
	if (!rc_avpair_add(rh, tosend, PW_AST_SIP_IP_ADDR, &cdr->sipIpAddress, strlen(cdr->sipIpAddress), VENDOR_CODE))
		return -1;

	/* Far End IP Address */
	if (!rc_avpair_add(rh, tosend, PW_AST_FAR_END_IP_ADDR, &cdr->farEndIPAddress, strlen(cdr->farEndIPAddress), VENDOR_CODE))
		return -1;

	/* Sip Response Code */
	if (!rc_avpair_add(rh, tosend, PW_AST_SIP_RESPONSE_CODE, &cdr->sipResponseCode, 0, VENDOR_CODE))
		return -1;

	/* codec */
	if (!rc_avpair_add(rh, tosend, PW_AST_CODEC, &cdr->codec, strlen(cdr->codec), VENDOR_CODE))
		return -1;

	/* RTP statistics */
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_LOCAL_BURST_DENSITY, &cdr->rtp_stats->localBurstDensity, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_REMOTE_BURST_DENSITY, &cdr->rtp_stats->remoteBurstDensity, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_LOCAL_BURST_DURATION, &cdr->rtp_stats->localBurstDuration, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_REMOTE_BURST_DURATION, &cdr->rtp_stats->remoteBurstDuration, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_LOCAL_GAP_DENSITY, &cdr->rtp_stats->localGapDensity, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_REMOTE_GAP_DENSITY, &cdr->rtp_stats->remoteGapDensity, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_LOCAL_GAP_DURATION, &cdr->rtp_stats->localGapDuration, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_REMOTE_GAP_DURATION, &cdr->rtp_stats->remoteGapDuration, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_LOCAL_JB_RATE, &cdr->rtp_stats->localJbRate, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_REMOTE_JB_RATE, &cdr->rtp_stats->remoteJbRate, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_LOCAL_JB_MAX, &cdr->rtp_stats->localJbMax, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_REMOTE_JB_MAX, &cdr->rtp_stats->remoteJbMax, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_LOCAL_JB_NOMINAL, &cdr->rtp_stats->localJbNominal, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_REMOTE_JB_NOMINAL, &cdr->rtp_stats->remoteJbNominal, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_LOCAL_JB_ABS_MAX, &cdr->rtp_stats->localJbAbsMax, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_REMOTE_JB_ABS_MAX, &cdr->rtp_stats->remoteJbAbsMax, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_JB_AVG, &cdr->rtp_stats->jbAvg, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_LOSS_RATE, &cdr->rtp_stats->uLossRate, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_DISCARDED, &cdr->rtp_stats->discarded, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_LOST, &cdr->rtp_stats->lost, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_RX_PKTS, &cdr->rtp_stats->rxpkts, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_TX_PKTS, &cdr->rtp_stats->txpkts, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_JITTER, &cdr->rtp_stats->jitter, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_MAX_JITTER, &cdr->rtp_stats->maxJitter, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_AVERAGE_ROUND_TRIP_DELAY, &cdr->rtp_stats->averageRoundTripDelay, 0, VENDOR_CODE))
		return -1;
	if (!rc_avpair_add(rh, tosend, PW_AST_RTP_AVERAGE_FAR_END_INTERARRIVAL_JITTER, &cdr->rtp_stats->averageFarEndInterarrivalJitter, 0, VENDOR_CODE))
		return -1;

	/* Setting Acct-Session-Id & User-Name attributes for proper generation
	 * of Acct-Unique-Session-Id on server side
	 */
	/* Channel */
	if (!rc_avpair_add(rh, tosend, PW_USER_NAME, &cdr->channel, strlen(cdr->channel), 0))
		return -1;

	/* Unique ID */
	if (!rc_avpair_add(rh, tosend, PW_ACCT_SESSION_ID, &cdr->uniqueid, strlen(cdr->uniqueid), 0))
		return -1;

	return 0;
}

static int radius_log(struct ast_cdr *cdr)
{
	int result = ERROR_RC;
	VALUE_PAIR *tosend = NULL;

	if (build_radius_record(&tosend, cdr)) {
		ast_debug(1, "Unable to create RADIUS record. CDR not recorded!\n");
		goto return_cleanup;
	}

	result = rc_acct(rh, 0, tosend);
	if (result != OK_RC) {
		ast_log(LOG_ERROR, "Failed to record Radius CDR record!\n");
	}

return_cleanup:
	if (tosend) {
		rc_avpair_free(tosend);
	}

	return result;
}

static int unload_module(void)
{
	if (ast_cdr_unregister(name)) {
		return -1;
	}

	if (rh) {
		rc_destroy(rh);
		rh = NULL;
	}
	return 0;
}

static int load_module(void)
{
	struct ast_config *cfg;
	struct ast_flags config_flags = { 0 };
	const char *tmp;

	if ((cfg = ast_config_load(cdr_config, config_flags)) && cfg != CONFIG_STATUS_FILEINVALID) {
		ast_set2_flag(&global_flags, ast_true(ast_variable_retrieve(cfg, "radius", "usegmtime")), RADIUS_FLAG_USEGMTIME);
		ast_set2_flag(&global_flags, ast_true(ast_variable_retrieve(cfg, "radius", "loguniqueid")), RADIUS_FLAG_LOGUNIQUEID);
		ast_set2_flag(&global_flags, ast_true(ast_variable_retrieve(cfg, "radius", "loguserfield")), RADIUS_FLAG_LOGUSERFIELD);
		if ((tmp = ast_variable_retrieve(cfg, "radius", "radiuscfg")))
			ast_copy_string(radiuscfg, tmp, sizeof(radiuscfg));
		ast_config_destroy(cfg);
	} else
		return AST_MODULE_LOAD_DECLINE;

	/* read radiusclient-ng config file */
	if (!(rh = rc_read_config(radiuscfg))) {
		ast_log(LOG_NOTICE, "Cannot load radiusclient-ng configuration file %s.\n", radiuscfg);
		return AST_MODULE_LOAD_DECLINE;
	}

	/* read radiusclient-ng dictionaries */
	if (rc_read_dictionary(rh, rc_conf_str(rh, "dictionary"))) {
		ast_log(LOG_NOTICE, "Cannot load radiusclient-ng dictionary file.\n");
		rc_destroy(rh);
		rh = NULL;
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_cdr_register(name, desc, radius_log)) {
		rc_destroy(rh);
		rh = NULL;
		return AST_MODULE_LOAD_DECLINE;
	} else {
		return AST_MODULE_LOAD_SUCCESS;
	}
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "RADIUS CDR Backend",
	.support_level = AST_MODULE_SUPPORT_EXTENDED,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_CDR_DRIVER,
	.requires = "cdr",
);
